import abc

import torch


class Instrument(abc.ABC):
    """
    Base class for all financial instruments.
    """

    @abc.abstractmethod
    def simulate(self, time_horizon, n_paths=1, init_price=1.0) -> None:
        """
        Simulate time series of prices of itself (for a primary instrument)
        or its underlier (for a derivative).
        """

    @abc.abstractmethod
    def to(self, *args, **kwargs):
        """
        Performs dtype and/or device conversion of the time series of prices.

        Parameters
        ----------
        - dtype : torch.dtype
            Desired floating point type of the floating point values
            of simulated time series.
        - device : torch.device
            Desired device of the values of simulated time series.

        Returns
        -------
        self
        """


class Primary(Instrument):
    """
    Base class for all primary instruments.

    A primary instrument is a basic financial instrument which is traded on a market
    and therefore the price is accessible as the market price.

    Examples include stocks, bonds, commodities, and currencies.

    Derivatives are issued based on primary instruments
    (See `Derivative` class for details).

    Attributes
    ----------
    - prices : Tensor, shape (N_STEPS, N_PATHS)
        The prices of the instrument.
        This attribute is supposed to set by a method `simulate()`.
        Here, `N_STEPS` is the number of time steps and
        `N_PATHS` is the number of simulated paths.
    """

    @abc.abstractmethod
    def simulate(self, time_horizon, n_paths=1, init_price=1.0, **kwargs) -> None:
        """
        Simulate time series of prices and set an attribute `prices`.

        Parameters
        ----------
        - time_horizon : float
            The period of time to simulate the price.
        - n_paths : int, default 1
            The number of paths to simulate.
        - init_price : float, default 1.0
            The initial value of the prices.
        """

    def to(self, *args, **kwargs):
        """
        Performs dtype and/or device conversion of the time series of the prices.

        Parameters
        ----------
        - dtype : torch.dtype
            Desired floating point type of the floating point values
            of simulated time series.
        - device : torch.device
            Desired device of the values of simulated time series.

        Returns
        -------
        self
        """
        device, dtype, *_ = torch._C._nn._parse_to(*args, **kwargs)

        if dtype is not None and not dtype.is_floating_point:
            raise TypeError(
                f"Instrument.to only accepts floating point "
                f"dtypes, but got desired dtype={dtype}"
            )

        if not hasattr(self, "dtype") or dtype is not None:
            self.dtype = dtype
        if not hasattr(self, "device") or device is not None:
            self.device = device

        # If prices have been already simulated, move it
        if hasattr(self, "prices"):
            self.prices = self.prices.to(*args, **kwargs)

        return self


class Derivative(Instrument):
    """
    Base class for all derivatives.

    A derivative is a financial instrument whose payoff is contingent on
    a primary instrument (or a set of primary instruments).
    A (over-the-counter) derivative is not traded on the market and therefore the price
    is not directly accessible.

    Examples include options and swaps.

    A derivative relies on primary assets (See `Primary` for details), such as
    stocks, bonds, commodities, and currencies.

    Attributes
    ----------
    - underlier : Primary
        The underlying asset on which the derivative's payoff relies.
    """

    @property
    def dtype(self) -> torch.dtype:
        return self.underlier.dtype

    @property
    def device(self) -> torch.device:
        return self.underlier.device

    def simulate(self, n_paths=1, init_price=1.0, **kwargs) -> None:
        """
        Simulates time series of the underlier's prices.

        Parameters
        ----------
        - n_paths : int
            The number of paths to simulate.
        - init_price : float
            The initial value of the prices.
        """
        self.underlier.simulate(
            time_horizon=self.maturity, n_paths=n_paths, init_price=init_price, **kwargs
        )

    def to(self, *args, **kwargs):
        """
        Performs dtype and/or device conversion of the underlier.

        Parameters
        ----------
        - dtype : torch.dtype
            Desired floating point type of the floating point values
            of simulated time series.
        - device : torch.device
            Desired device of the values of simulated time series.

        Returns
        -------
        self
        """
        self.underlier.to(*args, **kwargs)
        return self

    @abc.abstractmethod
    def payoff(self) -> torch.Tensor:
        """
        Returns the payoffs of the derivative.
        The payoffs is computed based on the prices of `underlier`.

        Returns
        -------
        payoff : Tensor, shape (N_PATHS,)
        """



import torch
from torch.optim import Adam
from tqdm import tqdm

from ._utils.hook import save_prev_output
from ._utils.operations import ensemble_mean
from .features import get_feature
from .nn import EntropicRiskMeasure


class Hedger(torch.nn.Module):
    """
    Module to hedge and price derivatives.

    Parameters
    ----------
    - model : torch.nn.Module
        Hedging model to compute the hedge ratio at the next time step from
        the input features at the current time step.
        The input and output shapes should be `(N, H_in) -> (N, 1)`,
        where `N` stands for the number of Monte Carlo paths of the asset prices
        and `H_in` stands for the number of input features
        (namely, `H_in = len(features)`).
    - features : list[str|Feature]
        List of (names of) features to feed to model.
        See `sorted(pfhedge.features.FEATURES)` for valid options.
    - criterion : HedgeLoss, default `EntropicRiskMeasure()`
        Loss function to minimize by hedging.

    Shape
    -----
    Returns the output of `model`.

    Input : (N, H_in)
        See `features` for input features.
        Here, `H_in` is the number of input features.
    Output : (N, 1)
        The hedge ratio at the next time step.

    Examples
    --------
    >>> from pfhedge.instruments import BrownianStock
    >>> from pfhedge.instruments import EuropeanOption

    >>> deriv = EuropeanOption(BrownianStock(cost=1e-4))

    Black-Scholes' delta hedging strategy.

    >>> from pfhedge.nn import BlackScholes

    >>> model = BlackScholes(deriv)
    >>> hedger = Hedger(model, model.features())
    >>> hedger
    Hedger(
      features=['log_moneyness', 'expiry_time', 'volatility'],
      (model): BSEuropeanOption()
      (criterion): EntropicRiskMeasure()
    )
    >>> _ = torch.manual_seed(42)
    >>> hedger.compute_pnl(deriv, n_paths=5)
    tensor([-0.0209, -0.0257, -0.0224, -0.0174, -0.0202])
    >>> _ = torch.manual_seed(42)
    >>> hedger.price(deriv)
    tensor(0.0223)

    Whalley-Wilmott's no-transaction-band strategy.

    >>> from pfhedge.nn import WhalleyWilmott

    >>> model = WhalleyWilmott(deriv)
    >>> hedger = Hedger(model, model.features())
    >>> hedger
    Hedger(
      features=['log_moneyness', 'expiry_time', 'volatility', 'prev_hedge'],
      (model): WhalleyWilmott(
        (bs): BSEuropeanOption()
        (clamp): Clamp()
      )
      (criterion): EntropicRiskMeasure()
    )
    >>> _ = torch.manual_seed(42)
    >>> hedger.compute_pnl(deriv, n_paths=5)
    tensor([-0.0229, -0.0357, -0.0221, -0.0062, -0.0251])
    >>> _ = torch.manual_seed(42)
    >>> hedger.price(deriv)
    tensor(0.0226)

    A naked position (never hedge at all).

    >>> from pfhedge.nn import Naked

    >>> hedger = Hedger(Naked(), ["zero"])
    >>> _ = torch.manual_seed(42)
    >>> hedger.compute_pnl(deriv, n_paths=5)
    tensor([-0.0259, -0.1227,  0.0000, -0.0090, -0.0541])
    >>> _ = torch.manual_seed(42)
    >>> hedger.price(deriv)
    tensor(0.0246)

    A strategy represented by a neural network (Deep Hedging).

    >>> from pfhedge.nn import MultiLayerPerceptron

    >>> _ = torch.manual_seed(42)
    >>> model = MultiLayerPerceptron()
    >>> hedger = Hedger(model, ["moneyness", "expiry_time", "volatility"])
    >>> _ = hedger.compute_pnl(deriv, n_paths=1)  # lazily derermine in_features
    >>> hedger
    Hedger(
      features=['moneyness', 'expiry_time', 'volatility'],
      (model): MultiLayerPerceptron(
        (0): Linear(in_features=3, out_features=32, bias=True)
        (1): ReLU()
        (2): Linear(in_features=32, out_features=32, bias=True)
        (3): ReLU()
        (4): Linear(in_features=32, out_features=32, bias=True)
        (5): ReLU()
        (6): Linear(in_features=32, out_features=32, bias=True)
        (7): ReLU()
        (8): Linear(in_features=32, out_features=1, bias=True)
        (9): Identity()
      )
      (criterion): EntropicRiskMeasure()
    )
    >>> _ = torch.manual_seed(42)
    >>> history = hedger.fit(deriv, verbose=False, n_paths=1, n_epochs=1)
    >>> _ = torch.manual_seed(42)
    >>> hedger.compute_pnl(deriv, n_paths=5)
    tensor([-0.0277, -0.1312,  0.0053, -0.0096, -0.0578], grad_fn=<SubBackward0>)
    >>> _ = torch.manual_seed(42)
    >>> hedger.price(deriv)
    tensor(0.0249)
    """

    def __init__(self, model, features, criterion=EntropicRiskMeasure()):
        super().__init__()

        self.model = model
        self.features = [get_feature(feature) for feature in features]
        self.criterion = criterion

        # This hook saves the hedger's previous output to an attribute `prev`.
        # The attribute `prev` may be referred to by the feature `PrevHedge`.
        self.register_forward_hook(save_prev_output)

    def forward(self, input):
        return self.model(input)

    def extra_repr(self) -> str:
        params = []
        if not isinstance(self.model, torch.nn.Module):
            params.append(f"model={self.model.__name__},")
        params.append(f"features={[str(f) for f in self.features]},")

        return "\n".join(params)

    def compute_pnl(self, derivative, n_paths=1000, init_price=1.0) -> torch.tensor:
        """
        Returns the profit and loss distribution after hedging.

        A hedger sells the derivative to its customer and obliges to settle the payoff
        at maturity. The dealer hedges the risk of this liability by trading
        the underlying instrument of the derivative based on `model`.
        The resulting profit and loss is obtained by adding up the payoff to the
        customer, capital gains from the underlying asset, and the transaction cost.

        Parameters
        ----------
        - derivative : Derivative
            The derivative to hedge.
        - n_paths : int, default 1000
            The number of simulated price paths of the underlying instrument.
        - init_price : float, default 1.0
            The initial price of the underlying instrument of the derivative.

        Returns
        -------
        profit_and_loss : Tensor, shape (n_paths,)
        """
        self.features = [feature.of(derivative, self) for feature in self.features]

        derivative.simulate(n_paths=n_paths, init_price=init_price)
        cashflow = derivative.underlier.prices[1:] - derivative.underlier.prices[:-1]

        self.prev = torch.zeros_like(derivative.underlier.prices[:1]).reshape(-1)
        pnl = 0

        # Simulate hedging over time.
        n_steps = derivative.underlier.prices.size()[0]
        for i in range(n_steps - 1):
            prev_hedge = self.prev.reshape(-1)

            # Compute the hedge ratio at the next time step.
            hedge = self(torch.cat([f[i] for f in self.features], 1)).reshape(-1)

            # Receive profit and loss from the underlying asset.
            pnl += hedge * cashflow[i]
            # Deduct transactoon cost.
            pnl -= (
                derivative.underlier.cost
                * torch.abs(hedge - prev_hedge)
                * derivative.underlier.prices[i]
            )

        # Settle the derivative's payoff.
        pnl -= derivative.payoff()

        # Delete the attribute `prev` in case a hedger has a feature `PrevHedge` and
        # one calls `compute_pnl` twice.
        # If `prev` is not deleted, `prev` at the last time step in the first call
        # would be referred to by `PrevHedge` at the first time step at the second call,
        # which results in an unexpected output (while we expect zeros).
        if hasattr(self, "prev"):
            delattr(self, "prev")

        return pnl

    def compute_loss(
        self, derivative, n_paths=1000, n_times=1, init_price=1.0, enable_grad=True
    ) -> torch.Tensor:
        """
        Returns the loss of the profit and loss distribution after hedging.

        Parameters
        ----------
        - derivative : Derivative
            The derivative to hedge.
        - n_paths : int, default 1000
            The number of simulated price paths of the underlying instrument.
        - n_times : int, default 1
            If `n_times > 1`, returns the ensemble mean of the losses computed
            through multiple simulations.
        - init_price : float, default 1.0
            The initial price of the underlying instrument of the derivative.
        - enable_grad : bool, default True
            Context-manager that sets gradient calculation to on or off.

        Returns
        -------
        loss : Tensor, shape ()
        """
        with torch.set_grad_enabled(enable_grad):
            loss = lambda: self.criterion(
                self.compute_pnl(derivative, init_price=init_price, n_paths=n_paths)
            )
            mean_loss = ensemble_mean(loss, n_times=n_times)

        return mean_loss

    def fit(
        self,
        derivative,
        n_epochs=100,
        n_paths=1000,
        n_times=1,
        optimizer=Adam,
        init_price=1.0,
        verbose=True,
    ):
        """
        Train the hedging model to hedge the given derivative.

        Parameters
        ----------
        - derivative : Derivative
            The derivative to hedge.
        - n_epochs : int, default 100
            Number of Monte-Carlo simulations.
        - n_paths : int, default 1000
            The number of simulated price paths of the underlying instrument.
        - n_times : int, default 1
            If `n_times > 1`, returns the ensemble mean of the losses computed
            through multiple simulations.
        - optimizer : torch.optim.Optimizer, default Adam
            The optimizer algorithm to use.
            It can be an instance or a class of `torch.optim.Optimizer`.
        - init_price : float, default 1.0
            The initial price of the underlying instrument of the derivative.
        - enable_grad : bool, default True
            Context-manager that sets gradient calculation to on or off.
        - verbose : bool, default True
            If `True`, print progress of the training to standard output.

        Returns
        -------
        history : list[float], len `n_epochs`
            Loss after each simulation.

        Examples
        --------
        >>> from pfhedge.instruments import BrownianStock
        >>> from pfhedge.instruments import EuropeanOption
        >>> from pfhedge.nn import MultiLayerPerceptron

        Using a custom optimizer:

        >>> from torch.optim import SGD

        >>> derivative = EuropeanOption(BrownianStock())
        >>> hedger = Hedger(MultiLayerPerceptron(), ["zero"])
        >>> # Run a "dummy" forward to initialize lazy parameters
        >>> _ = hedger.compute_pnl(derivative, n_paths=1)
        >>> _ = hedger.fit(
        ...     derivative,
        ...     optimizer=SGD(hedger.model.parameters(), lr=0.1),
        ...     n_epochs=1,
        ...     verbose=False,
        ... )

        You can also pass a class object of an optimizer.

        >>> from torch.optim import Adadelta

        >>> derivative = EuropeanOption(BrownianStock())
        >>> hedger = Hedger(MultiLayerPerceptron(), ["zero"])
        >>> _ = hedger.fit(
        ...     derivative,
        ...     optimizer=Adadelta,
        ...     n_epochs=1,
        ...     verbose=False,
        ... )
        """
        if isinstance(optimizer, type):
            # Run a "dummy" forward to initialize lazy parameters
            _ = self.compute_pnl(derivative, n_paths=1)
            optimizer = optimizer(self.model.parameters())
            if not isinstance(optimizer, torch.optim.Optimizer):
                raise TypeError("optimizer is not torch.optim.Optimizer")

        compute_loss = lambda **kwargs: self.compute_loss(
            derivative, n_paths=n_paths, init_price=init_price, **kwargs
        )

        history = []
        progress = tqdm(range(n_epochs)) if verbose else range(n_epochs)
        for _ in progress:
            # Compute training loss and backpropagate
            self.train()
            optimizer.zero_grad()
            loss = compute_loss()
            loss.backward()
            optimizer.step()

            # Compute validation loss
            self.eval()
            loss = compute_loss(n_times=n_times, enable_grad=False)
            history.append(loss.item())

            if verbose:
                progress.desc = f"Loss={loss:.5e}"

        return history

    def price(
        self, derivative, n_paths=1000, n_times=1, init_price=1.0, enable_grad=False
    ) -> torch.Tensor:
        """
        Evaluate the premium of the given derivative.

        Parameters
        ----------
        - derivative : Derivative
            The derivative to price.
        - n_paths : int, default 1000
            The number of simulated price paths of the underlying instrument.
        - n_times : int, default 1
            If `n_times > 1`, returns the ensemble mean of the losses computed
            through multiple simulations.
        - init_price : float, default 1.0
            The initial price of the underlying instrument of the derivative.
        - enable_grad : bool, default False
            Context-manager that sets gradient calculation to on or off.

        Returns
        -------
        price : Tensor, shape ()
        """
        with torch.set_grad_enabled(enable_grad):
            # Negative because selling
            price = lambda: -self.criterion.cash(
                self.compute_pnl(derivative, init_price=init_price, n_paths=n_paths)
            )
            mean_price = ensemble_mean(price, n_times=n_times)

        return mean_price



from ..nn.functional import american_binary_payoff
from ._base import Derivative


class AmericanBinaryOption(Derivative):
    """
    An American binary Option.

    An American binary call option pays an unit amount of cash if and only if
    the maximum of the underlying asset's price until maturity is equal or greater
    than the strike price.

    An American binary put option pays an unit amount of cash if and only if
    the minimum of the underlying asset's price until maturity is equal or smaller
    than the strike price.

    The payoff of an American binary call option is given by:

        payoff = 1 if Max >= K
                 0 otherwise

        Max = The maximum of the underlying asset's price until maturity
        K = The strike price (`strike`) of the option

    The payoff of an American binary put option is given by:

        payoff = 1 if Min <= K
                 0 otherwise

        Min = The minimum of the underlying asset's price until maturity
        K = The strike price (`strike`) of the option

    Parameters
    ----------
    - underlier : Primary
        The underlying instrument of the option.
    - call : bool, default True
        Specify whether the option is call or put.
    - strike : float, default 1.0
        The strike price of the option.
    - maturity : float, default 30 / 365
        The maturity of the option.

    Examples
    --------
    >>> import torch
    >>> from pfhedge.instruments import BrownianStock

    >>> _ = torch.manual_seed(42)
    >>> deriv = AmericanBinaryOption(
    ...     BrownianStock(), maturity=5 / 365, strike=1.01)
    >>> deriv.simulate(n_paths=2)
    >>> deriv.underlier.prices
    tensor([[1.0000, 1.0000],
            [1.0024, 1.0024],
            [0.9906, 1.0004],
            [1.0137, 0.9936],
            [1.0186, 0.9964]])
    >>> deriv.payoff()
    tensor([1., 0.])
    """

    def __init__(self, underlier, call=True, strike=1.0, maturity=30 / 365):
        super().__init__()
        self.underlier = underlier
        self.call = call
        self.strike = strike
        self.maturity = maturity

    def __repr__(self):
        params = [f"{self.underlier.__class__.__name__}(...)"]
        if not self.call:
            params.append(f"call={self.call}")
        if self.strike != 1.0:
            params.append(f"strike={self.strike}")
        params.append(f"maturity={self.maturity:.2e}")

        return self.__class__.__name__ + f"({', '.join(params)})"

    def payoff(self):
        return american_binary_payoff(
            self.underlier.prices, call=self.call, strike=self.strike
        )



import torch

from ..._utils.bisect import bisect
from ._base import BSModuleMixin


class BSEuropeanBinaryOption(BSModuleMixin):
    """
    Black-Scholes formula for a European binary option.

    Parameters
    ----------
    - derivative : EuropeanBinaryOption
        The derivative to get the Black-Scholes formula.
    - call : bool, default True
        Specify whether the option is call or put.
    - strike : float, default 1.0
        The strike price of the option.

    Shape
    -----
    - Input : (N, *, 3)
        See `features()` for input features.
        Here, `*` means any number of additional dimensions.
    - Output : (N, *, 1)
        Delta of the derivative.
        All but the last dimension are the same shape as the input.

    Examples
    --------
    >>> m = BSEuropeanBinaryOption(strike=1.0)
    >>> m.features()
    ['log_moneyness', 'expiry_time', 'volatility']
    >>> x = torch.tensor([
    ...     [-0.01, 0.1, 0.2],
    ...     [ 0.00, 0.1, 0.2],
    ...     [ 0.01, 0.1, 0.2]])
    >>> m(x)
    tensor([[6.2576],
            [6.3047],
            [6.1953]])
    """

    def __init__(self, derivative=None, call=True, strike=1.0):
        super().__init__()

        if derivative is not None:
            self.call = derivative.call
            self.strike = derivative.strike
        else:
            self.call = call
            self.strike = strike

        if not self.call:
            raise ValueError(
                f"{self.__class__.__name__} for a put option is not yet supported."
            )

    def extra_repr(self):
        params = []
        if self.strike != 1.0:
            params.append(f"strike={self.strike}")
        return ", ".join(params)

    def features(self):
        return ["log_moneyness", "expiry_time", "volatility"]

    @torch.enable_grad()
    def delta(
        self, log_moneyness, expiry_time, volatility, create_graph=False
    ) -> torch.Tensor:
        """
        Returns delta of the derivative.

        Parameters
        ----------
        - log_moneyness : Tensor, shape (N, *)
            Log moneyness of the prices of the underlying asset.
        - expiry_time : Tensor, shape (N, *)
            Time to expiry of the option.
        - volatility : Tensor, shape (N, *)
            Volatility of the underlying asset.
        - create_graph : bool, default False
            If True, graph of the derivative will be constructed.
            This option is used to compute gamma.

        Returns
        -------
        delta : Tensor, shape (N, *)
        """
        s, t, v = map(torch.as_tensor, (log_moneyness, expiry_time, volatility))

        delta = self.N.pdf(self.d2(s, t, v)) / (
            self.strike * torch.exp(s) * v * torch.sqrt(t)
        )
        return delta

    def gamma(self, log_moneyness, expiry_time, volatility) -> torch.Tensor:
        """
        Returns gamma of the derivative.

        Parameters
        ----------
        - log_moneyness : Tensor, shape (N, *)
            Log moneyness of the prices of the underlying asset.
        - expiry_time : Tensor, shape (N, *)
            Time to expiry of the option.
        - volatility : Tensor, shape (N, *)
            Volatility of the underlying asset.

        Returns
        -------
        gamma : Tensor, shape (N, *)
        """
        prices = self.strike * torch.exp(torch.as_tensor(log_moneyness))
        prices = torch.Tensor.requires_grad_(prices)

        s = torch.log(prices / self.strike)
        t, v = map(torch.as_tensor, (expiry_time, volatility))

        delta = self.delta(s, t, v, create_graph=True)
        gamma = torch.autograd.grad(
            delta, prices, grad_outputs=torch.ones_like(prices)
        )[0]

        return gamma

    def price(self, log_moneyness, expiry_time, volatility) -> torch.Tensor:
        """
        Returns price of the derivative.

        Parameters
        ----------
        - log_moneyness : Tensor, shape (N, *)
            Log moneyness of the prices of the underlying asset.
        - expiry_time : Tensor, shape (N, *)
            Time to expiry of the option.
        - volatility : Tensor, shape (N, *)
            Volatility of the underlying asset.

        Returns
        -------
        price : Tensor, shape (N, *)
        """
        s, t, v = map(torch.as_tensor, (log_moneyness, expiry_time, volatility))

        price = self.N.cdf(self.d2(s, t, v))
        price = 1.0 - price if not self.call else price  # put-call parity

        return price

    def implied_volatility(
        self, log_moneyness, expiry_time, price, precision=1e-6
    ) -> torch.Tensor:
        """
        Returns implied volatility of the derivative.

        Parameters
        ----------
        - log_moneyness : Tensor, shape (N, *)
            Log moneyness of the prices of the underlying asset.
        - expiry_time : Tensor, shape (N, *)
            Time to expiry of the option.
        - price : Tensor, shape (N, *)
            Price of the derivative.
        - precision : float, default 1e-6
            Computational precision of the implied volatility.

        Returns
        -------
        implied_volatility : Tensor, shape (N, *)
        """
        s, t, p = map(torch.as_tensor, (log_moneyness, expiry_time, price))
        get_price = lambda v: self.price(s, t, v)
        return bisect(get_price, p, lower=0.001, upper=1.000, precision=precision)



import abc

import torch


class Feature(abc.ABC):
    """
    Base class for all features.

    All features should subclass this class.
    """

    @abc.abstractmethod
    def __getitem__(self, i) -> torch.Tensor:
        """
        Return feature tensor.

        Parameters
        ----------
        i : int
            The index of the time step to get the feature.

        Returns
        -------
        feature : Tensor, shape (n_paths, 1)
            where `n_paths` is the number of simulated price paths.
        """

    def of(self, derivative=None, hedger=None):
        """
        Set `derivative` and `hedger` to the attributes of `self`.

        Parameters
        ----------
        - derivative : Derivative, optional
            The derivative to compute features.
        - hedger : Hedger, optional
            The hedger to compute features.

        Returns
        -------
        self
        """
        if not hasattr(self, "derivative") or derivative is not None:
            self.derivative = derivative
        if not hasattr(self, "hedger") or derivative is not None:
            self.hedger = hedger
        return self



def save_prev_output(module, input, output) -> None:
    """
    Save previous output as an attribute `prev`.

    Parameters
    ----------
    - module : torch.nn.Module
        The module.
    - input : Tensor
        The input tensor.
    - output : Tensor
        The output tensor.

    Examples
    --------
    >>> import torch
    >>> _ = torch.manual_seed(42)
    >>> m = torch.nn.Linear(3, 2)
    >>> hook = m.register_forward_hook(save_prev_output)
    >>> x = torch.randn(1, 3)
    >>> m(x)
    tensor([[-1.1647,  0.0244]], grad_fn=<AddmmBackward>)
    >>> m.prev
    tensor([[-1.1647,  0.0244]], grad_fn=<AddmmBackward>)
    """
    module.prev = output



import torch


def generate_brownian(
    n_steps,
    n_paths,
    init_value=0.0,
    volatility=0.2,
    dt=1.0 / 365,
    dtype=None,
    device=None,
) -> torch.Tensor:
    """
    Returns time series following the Brownian motion.
    The drift of the time series is assumed to be vanishing.

    Parameters
    ----------
    - n_steps : int
        The number of time steps.
    - n_paths : int
        The number of simulated paths.
    - init_value : float, default 0.0
        The initial value of the time series.
    - volatility : float, default 0.2
        The volatility of the Brownian motion.
    - dt : float, default 1.0 / 365
        The intervals of the time steps.
    - dtype : torch.dtype, optional
        The desired data type of returned tensor.
        Default: If `None`, uses a global default
        (see `torch.set_default_tensor_type()`).
    - device : torch.device, optional
        The desired device of returned tensor.
        Default: if None, uses the current device for the default tensor type
        (see `torch.set_default_tensor_type()`).
        `device` will be the CPU for CPU tensor types and the current CUDA device
        for CUDA tensor types.

    Returns
    -------
    brownian : Tensor, shape (n_steps, n_paths)

    Examples
    --------
    >>> _ = torch.manual_seed(42)
    >>> generate_brownian(5, 2)
    tensor([[ 0.0000,  0.0000],
            [ 0.0025,  0.0024],
            [-0.0093,  0.0005],
            [ 0.0138, -0.0062],
            [ 0.0186, -0.0034]])
    """
    randn = torch.randn((n_steps, n_paths), dtype=dtype, device=device)
    randn[0] = 0.0
    return init_value + volatility * torch.tensor(dt).to(randn).sqrt() * randn.cumsum(0)


def generate_geometric_brownian(
    n_steps,
    n_paths,
    init_value=1.0,
    volatility=0.2,
    dt=1.0 / 365,
    dtype=None,
    device=None,
) -> torch.Tensor:
    """
    Returns time series following the geometric Brownian motion.
    The drift of the time series is assumed to be vanishing.

    Parameters
    ----------
    - n_steps : int
        The number of time steps.
    - n_paths : int
        The number of simulated paths.
    - init_value : float, default 0.0
        The initial value of the time series.
    - volatility : float, default 0.2
        The volatility of the Brownian motion.
    - dt : float, default 1.0 / 365
        The intervals of the time steps.
    - dtype : torch.dtype, optional
        The desired data type of returned tensor.
        Default: If `None`, uses a global default
        (see `torch.set_default_tensor_type()`).
    - device : torch.device, optional
        The desired device of returned tensor.
        Default: if None, uses the current device for the default tensor type
        (see `torch.set_default_tensor_type()`).
        `device` will be the CPU for CPU tensor types and the current CUDA device
        for CUDA tensor types.

    Returns
    -------
    geometric_brownian : Tensor, shape (n_steps, n_paths)

    Examples
    --------
    >>> _ = torch.manual_seed(42)
    >>> generate_geometric_brownian(5, 2)
    tensor([[1.0000, 1.0000],
            [1.0024, 1.0024],
            [0.9906, 1.0004],
            [1.0137, 0.9936],
            [1.0186, 0.9964]])
    """
    brownian = generate_brownian(
        n_steps=n_steps,
        n_paths=n_paths,
        init_value=0.0,
        volatility=volatility,
        dt=dt,
        dtype=dtype,
        device=device,
    )
    t = dt * torch.arange(n_steps).to(brownian).reshape(-1, 1)
    return init_value * torch.exp(brownian - (volatility ** 2) * t / 2)



import torch


class Naked(torch.nn.Module):
    """
    Returns a tensor filled with the scalar value zero.

    Parameters
    ----------
    - out_features : int, default 1
        Size of each output sample.

    Shape
    -----
    - Input : (N, *, H_in)
        where where * means any number of
        additional dimensions and `H_in = in_features`.
    - Output : (N, *, H_out)
        where all but the last dimension
        are the same shape as the input and `H_out = out_features`.

    Examples
    --------
    >>> m = Naked()
    >>> input = torch.empty((2, 3))
    >>> m(input)
    tensor([[0.],
            [0.]])
    """

    def __init__(self, out_features=1):
        super().__init__()
        self.out_features = out_features

    def forward(self, input):
        return torch.cat(
            [torch.zeros_like(input[..., :1]) for _ in range(self.out_features)], -1
        )



import torch


def ensemble_mean(function, n_times=1, *args, **kwargs) -> torch.Tensor:
    """
    Compute ensemble mean from function.

    Parameters
    ----------
    - function : callable[..., Tensor], shape (*)
        Function to evaluate.
    - n_times : int, default 1
        Number of times to evaluate.
    - *args, **kwargs
        Arguments passed to the function.

    Returns
    -------
    ensemble_mean : Tensor, shape (*)

    Examples
    --------
    >>> function = lambda: torch.tensor([1.0, 2.0])
    >>> ensemble_mean(function, 5)
    tensor([1., 2.])

    >>> _ = torch.manual_seed(42)
    >>> function = lambda: torch.randn(2)
    >>> ensemble_mean(function, 5)
    tensor([ 0.4236, -0.0396])
    """
    return torch.stack([function(*args, **kwargs) for _ in range(n_times)]).mean(dim=0)



import torch

from ...instruments import AmericanBinaryOption
from ...instruments import EuropeanBinaryOption
from ...instruments import EuropeanOption
from ...instruments import LookbackOption
from .american_binary import BSAmericanBinaryOption
from .european import BSEuropeanOption
from .european_binary import BSEuropeanBinaryOption
from .lookback import BSLookbackOption


class BlackScholes(torch.nn.Module):
    """
    Initialize Black-Scholes formula module from a derivative.
    The `forward` method returns the Black-Scholes delta.

    Parameters
    ----------
    - derivative : Derivative
        The derivative to get the Black-Scholes formula.

    Shape
    -----
    - Input : (N, *, H_in)
        See `features()` for input features.
        Here, `*` means any number of additional dimensions and `H_in` is
        the number of input features..
    - Output : (N, *, 1)
        where all but the last dimension are the same shape as the input.

    Examples
    --------
    >>> import torch
    >>> from pfhedge.instruments import BrownianStock
    >>> from pfhedge.instruments import EuropeanOption
    >>> from pfhedge.instruments import LookbackOption

    >>> deriv = EuropeanOption(BrownianStock())
    >>> m = BlackScholes(deriv)
    >>> m
    BSEuropeanOption()
    >>> m.features()
    ['log_moneyness', 'expiry_time', 'volatility']
    >>> x = torch.tensor([
    ...     [-0.01, 0.1, 0.2],
    ...     [ 0.00, 0.1, 0.2],
    ...     [ 0.01, 0.1, 0.2]])
    >>> m(x)
    tensor([[0.4497],
            [0.5126],
            [0.5752]])

    >>> deriv = LookbackOption(BrownianStock(), strike=1.03)
    >>> m = BlackScholes(deriv)
    >>> m
    BSLookbackOption(strike=1.03)
    >>> m.features()
    ['log_moneyness', 'max_log_moneyness', 'expiry_time', 'volatility']
    >>> x = torch.tensor([
    ...     [-0.01, -0.01, 0.1, 0.2],
    ...     [ 0.00,  0.00, 0.1, 0.2],
    ...     [ 0.01,  0.01, 0.1, 0.2]])
    >>> m(x)
    tensor([[0.9208],
            [1.0515],
            [1.0515]])
    """

    def __init__(self, derivative):
        self.__class__ = {
            EuropeanOption: BSEuropeanOption,
            LookbackOption: BSLookbackOption,
            AmericanBinaryOption: BSAmericanBinaryOption,
            EuropeanBinaryOption: BSEuropeanBinaryOption,
        }[derivative.__class__]
        self.__init__(derivative)



import torch

from ..functional import clamp
from ..functional import leaky_clamp


class LeakyClamp(torch.nn.Module):
    """
    Clamp all elements in the input tensor into the range [`min_value`, `max_value`]
    and return a resulting tensor.
    The bounds `min_value` and `max_value` can be tensors.

    If min_value <= max_value:

        output = min_value + clamped_slope * (input - min_value) if input < min_value
                 input if min_value <= input <= max_value
                 max_value + clamped_slope * (input - max_value) if input > max_value

    If min_value > max_value:

        output = (min_value + max_value) / 2

    Parameters
    ----------
    - clamped_slope : float, default 0.01
        Controls the slope in the clampled regions.

    Shape
    -----
    Inputs: input, min_value, max_value
    Output: output
    - input : Tensor, shape (*)
        The input tensor.
    - min_value : float or Tensor, shape (*), optional
        Lower-bound of the range to be clamped to.
    - max_value : float or Tensor, shape (*), optional
        Upper-bound of the range to be clamped to.
    - output : Tensor, shape (*)
        The output tensor.

    Examples
    --------
    >>> m = LeakyClamp()
    >>> input = torch.linspace(-2, 12, 15) * 0.1
    >>> input
    tensor([-0.2000, -0.1000,  0.0000,  0.1000,  0.2000,  0.3000,  0.4000,  0.5000,
             0.6000,  0.7000,  0.8000,  0.9000,  1.0000,  1.1000,  1.2000])
    >>> m(input, 0.0, 1.0)
    tensor([-2.0000e-03, -1.0000e-03,  0.0000e+00,  1.0000e-01,  2.0000e-01,
             3.0000e-01,  4.0000e-01,  5.0000e-01,  6.0000e-01,  7.0000e-01,
             8.0000e-01,  9.0000e-01,  1.0000e+00,  1.0010e+00,  1.0020e+00])
    """

    def __init__(self, clamped_slope=0.01):
        super().__init__()
        self.clamped_slope = clamped_slope

    def extra_repr(self):
        return f"clamped_slope={self.clamped_slope}" if self.clamped_slope != 0 else ""

    def forward(self, input, min_value=None, max_value=None) -> torch.Tensor:
        return leaky_clamp(
            input, min_value, max_value, clamped_slope=self.clamped_slope
        )


class Clamp(torch.nn.Module):
    """
    Clamp all elements in the input tensor into the range [`min_value`, `max_value`]
    and return a resulting tensor.
    The bounds `min_value` and `max_value` can be tensors.

    If min_value <= max_value:

        output = min_value if input < min_value
                 input if min_value <= input <= max_value
                 max_value if input > max_value

    If min_value > max_value:

        output = (min_value + max_value) / 2

    Shape
    -----
    Inputs: input, min_value, max_value
    Output: output
    - input : Tensor, shape (*)
        The input tensor.
    - min_value : float or Tensor, shape (*), optional
        Lower-bound of the range to be clamped to.
    - max_value : float or Tensor, shape (*), optional
        Upper-bound of the range to be clamped to.
    - output : Tensor, shape (*)
        The output tensor.

    Returns
    -------
    output : Tensor, shape (*)
        The output tensor.

    Examples
    --------
    >>> m = Clamp()
    >>> input = torch.linspace(-2, 12, 15) * 0.1
    >>> input
    tensor([-0.2000, -0.1000,  0.0000,  0.1000,  0.2000,  0.3000,  0.4000,  0.5000,
             0.6000,  0.7000,  0.8000,  0.9000,  1.0000,  1.1000,  1.2000])
    >>> m(input, 0.0, 1.0)
    tensor([0.0000, 0.0000, 0.0000, 0.1000, 0.2000, 0.3000, 0.4000, 0.5000, 0.6000,
            0.7000, 0.8000, 0.9000, 1.0000, 1.0000, 1.0000])

    >>> input = torch.tensor([1.0, 0.0])
    >>> m(input, [0.0, 1.0], 0.0)
    tensor([0.0000, 0.5000])
    """

    def forward(self, input, min_value=None, max_value=None):
        return clamp(input, min_value, max_value)



from abc import ABC

import torch
from torch.nn import Parameter

from ..._utils.bisect import bisect
from ..functional import exp_utility
from ..functional import expected_shortfall
from ..functional import isoelastic_utility


class HedgeLoss(torch.nn.Module, ABC):
    """
    Base class for hedging criteria.

    Shape
    -----
    Input : (N, *)
        Profit and loss distribution.
        Here, `*` means any number of additional dimensions.
    Output : (*)
        Loss.
    """

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        """
        Returns the loss of the profit-loss distribution.
        This method should be overridden.

        Parameters
        ----------
        - input : Tensor, shape (N, *)
            The distribution of the profit and loss.
            Here, `*` means any number of additional dimensions.

        Returns
        -------
        loss : Tensor, shape (*)
        """

    def cash(self, input: torch.Tensor) -> torch.Tensor:
        """
        Returns the cash amount which is as preferable as
        the given profit-loss distribution in terms of the loss.

        The output `cash` is expected to satisfy the following relation:

            loss(torch.full_like(pnl, cash)) = loss(pnl)

        By default, the output is computed by binary search.
        If analytic form is known, it is recommended to override this method
        for faster computation.

        Parameters
        ----------
        - input : Tensor, shape (N, *)
            The distribution of the profit and loss.
            Here, `*` means any number of additional dimensions.

        Returns
        -------
        cash : Tensor, shape (*)
        """
        return bisect(self, self(input), torch.min(input), torch.max(input))


class EntropicRiskMeasure(HedgeLoss):
    """
    Creates a loss given by the entropic risk measure.

    The entropic risk measure of the profit-loss distribution `pnl` is given by

        loss(pnl) = log(-E[u(pnl)]) / a
        u(x) = -exp(-a * x)

    Parameters
    ----------
    - a : float > 0, default 1.0
        Risk aversion coefficient of the exponential utility.

    Shape
    -----
    Input : (N, *)
        The distribution of the profit and loss.
        Here, `*` means any number of additional dimensions.
    Output : (*)
        Loss.

    Shape
    -----
    >>> loss = EntropicRiskMeasure()
    >>> pnl = -torch.arange(4.0)
    >>> loss(pnl)
    tensor(2.0539)
    >>> loss.cash(pnl)
    tensor(-2.0539)
    """

    def __init__(self, a=1.0):
        if not a > 0:
            raise ValueError("Risk aversion coefficient should be positive.")

        super().__init__()
        self.a = a

    def extra_repr(self):
        return f"a={self.a}" if self.a != 1 else ""

    def forward(self, input):
        return torch.log(-torch.mean(exp_utility(input, a=self.a), dim=0)) / self.a

    def cash(self, input):
        """
        Returns the cash amount which is as preferable as
        the given profit-loss distribution in terms of the loss.

        Parameters
        ----------
        - input : Tensor, shape (N, *)
            The distribution of the profit and loss.
            Here, `*` means any number of additional dimensions.

        Returns
        -------
        cash : Tensor, shape (*)
        """
        return -self(input)


class EntropicLoss(HedgeLoss):
    """
    Creates a loss given by the negative of expected exponential utility.

    The loss of the profit-loss `pnl` is given by

        loss(pnl) = -E[u(pnl)]
        u(x) = -exp(-a * x)

    Parameters
    ----------
    - a : float > 0, default 1.0
        Risk aversion coefficient of the exponential utility.

    Shape
    -----
    Input : (N, *)
        The distribution of the profit and loss.
        Here, `*` means any number of additional dimensions.
    Output : (*)
        Loss.

    Examples
    --------
    >>> loss = EntropicLoss()
    >>> pnl = -torch.arange(4.0)
    >>> loss(pnl)
    tensor(7.7982)
    >>> loss.cash(pnl)
    tensor(-2.0539)
    """

    def __init__(self, a=1.0):
        if not a > 0:
            raise ValueError("Risk aversion coefficient should be positive.")

        super().__init__()
        self.a = a

    def extra_repr(self):
        return f"a={self.a}" if self.a != 1 else ""

    def forward(self, input):
        return -torch.mean(exp_utility(input, a=self.a), dim=0)

    def cash(self, input):
        """
        Returns the cash amount which is as preferable as
        the given profit-loss distribution in terms of the loss.

        Parameters
        ----------
        - input : Tensor, shape (N, *)
            The distribution of the profit and loss.
            Here, `*` means any number of additional dimensions.

        Returns
        -------
        cash : Tensor, shape (*)
        """
        return -torch.log(-torch.mean(exp_utility(input, a=self.a), dim=0)) / self.a


class IsoelasticLoss(HedgeLoss):
    """
    Creates a loss function that measures the isoelastic utility.

    The loss of the profit-loss distribution `pnl` is given by

        loss(pnl) = -E[u(pnl)]

        u(x) = x ** (1 - a) if a != 1
               log(x) if a == 1

    Parameters
    ----------
    - a : float
        Relative risk aversion coefficient of the isoelastic utility.
        This parameter should satisfy `0 < a <= 1`.

    Shape
    -----
    Input : (N, *)
        The distribution of the profit and loss.
        Here, `*` means any number of additional dimensions.
    Output : (*)
        Loss.

    Examples
    --------
    >>> loss = IsoelasticLoss(0.5)
    >>> pnl = torch.arange(1.0, 5.0)
    >>> loss(pnl)
    tensor(-1.5366)
    >>> loss.cash(pnl)
    tensor(2.3610)

    >>> loss = IsoelasticLoss(1.0)
    >>> pnl = torch.arange(1.0, 5.0)
    >>> loss(pnl)
    tensor(-0.7945)
    >>> loss.cash(pnl)
    tensor(2.2134)
    """

    def __init__(self, a):
        if not 0 < a <= 1:
            raise ValueError(
                "Relative risk aversion coefficient should satisfy 0 < a <= 1."
            )

        super().__init__()
        self.a = a

    def extra_repr(self):
        return f"a={self.a}"

    def forward(self, input):
        return -torch.mean(isoelastic_utility(input, a=self.a), dim=0)


class ExpectedShortfall(HedgeLoss):
    """
    Creates a criterion that measures the expected shortfall.

    Parameters
    ----------
    - p : float in (0.0, 1.0], default 0.1
        Quantile level.
        This parameter should satisfy 0 < p <= 1.

    Shape
    -----
    Input : (N, *)
        The distribution of the profit and loss.
        Here, `*` means any number of additional dimensions.
    Output : (*)
        Loss.

    Examples
    --------
    >>> loss = ExpectedShortfall(0.5)
    >>> pnl = -torch.arange(4.0)
    >>> loss(pnl)
    tensor(2.5000)
    >>> loss.cash(pnl)
    tensor(-2.5000)
    """

    def __init__(self, p=0.1):
        if not 0 < p <= 1:
            raise ValueError("The quantile level should satisfy 0 < p <= 1.")

        super().__init__()
        self.p = p

    def extra_repr(self):
        return str(self.p)

    def forward(self, input):
        return expected_shortfall(input, p=self.p, dim=0)

    def cash(self, input):
        return -self(input)


class OCE(HedgeLoss):
    """
    Creates a criterion that measures the optimized certainty equivalent.

    The certainty equivalent is given by:

        loss(X, w) = w - E[u(X + w)]

    Minimization of loss gives the optimized certainty equivalent.

        rho_u(X) = inf_w loss(X, w)

    Parameters
    ----------
    - utility : torch.autograd.Function
        Utility function.

    Attributes
    ----------
    - w : torch.nn.Parameter
        Wealth.

    Examples
    --------
    >>> _ = torch.manual_seed(42)
    >>> m = OCE(lambda x: 1 - torch.exp(-x))
    >>> pnl = torch.randn(10)
    >>> m(pnl)
    tensor(0.0855, grad_fn=<SubBackward0>)
    >>> m.cash(pnl)
    tensor(-0.0821)
    """

    def __init__(self, utility):
        super().__init__()

        self.utility = utility
        self.w = Parameter(torch.tensor(0.0))

    def extra_repr(self):
        return self.utility.__name__ + f", w={self.w}"

    def forward(self, input):
        return self.w - torch.mean(self.utility(input + self.w), dim=0)



from ..stochastic import generate_geometric_brownian
from ._base import Primary


class BrownianStock(Primary):
    """
    A stock of which prices follow the geometric Brownian motion.
    The drift of the prices is assumed to be vanishing.

    Parameters
    ----------
    - volatility : float, default 0.20
        The volatility of the price.
    - cost : float, default 0.0
        The transaction cost rate.
    - dt : float, default 1.0 / 365
        The intervals of the time steps.
    - dtype : torch.device, optional
        Desired device of returned tensor.
        Default: If None, uses a global default (see `torch.set_default_tensor_type()`).
    - device : torch.device, optional
        Desired device of returned tensor.
        Default: if None, uses the current device for the default tensor type
        (see `torch.set_default_tensor_type()`).
        `device` will be the CPU for CPU tensor types and
        the current CUDA device for CUDA tensor types.

    Attributes
    ----------
    - prices : Tensor, shape (N_STEPS, N_PATHS)
        The prices of the instrument.
        This attribute is supposed to set by a method `simulate()`.
        Here, `N_STEPS` is the number of time steps and
        `N_PATHS` is the number of simulated paths.

    Examples
    --------
    >>> import torch

    >>> _ = torch.manual_seed(42)
    >>> stock = BrownianStock(volatility=0.20)
    >>> stock.simulate(time_horizon=5 / 365, n_paths=2)
    >>> stock.prices
    tensor([[1.0000, 1.0000],
            [1.0024, 1.0024],
            [0.9906, 1.0004],
            [1.0137, 0.9936],
            [1.0186, 0.9964]])
    """

    def __init__(self, volatility=0.2, cost=0.0, dt=1.0 / 365, dtype=None, device=None):
        super().__init__()

        self.volatility = volatility
        self.cost = cost
        self.dt = dt

        self.to(dtype=dtype, device=device)

    def __repr__(self):
        params = [f"volatility={self.volatility:.2e}"]

        if self.cost != 0.0:
            params.append(f"cost={self.cost:.2e}")
        params.append(f"dt={self.dt:.2e}")
        if hasattr(self, "dtype") and self.dtype is not None:
            params.append(f"dtype={self.dtype}")
        if hasattr(self, "device") and self.device is not None:
            params.append(f"device={self.device}")

        return self.__class__.__name__ + f"({', '.join(params)})"

    def simulate(self, time_horizon, n_paths=1, init_price=1.0) -> None:
        """
        Simulate time series of prices and set an attribute `prices`.

        Parameters
        ----------
        - time_horizon : float
            The period of time to simulate the price.
        - n_paths : int, default 1
            The number of paths to simulate.
        - init_price : float, default 1.0
            The initial value of the prices.
        """
        n_steps = int(time_horizon / self.dt)
        self.prices = generate_geometric_brownian(
            n_steps=n_steps,
            n_paths=n_paths,
            init_value=init_price,
            volatility=self.volatility,
            dt=self.dt,
            dtype=self.dtype,
            device=self.device,
        )



from copy import deepcopy

import torch
from torch.nn import Identity
from torch.nn import LazyLinear
from torch.nn import ReLU


class MultiLayerPerceptron(torch.nn.ModuleList):
    """
    A feed-forward neural network.

    Number of input features is lazily determined.

    Parameters
    ----------
    - out_features : int, default 1
        Size of each output sample.
    - n_layers : int, default 4
        Number of hidden layers.
    - n_units : int, default 32
        Number of units in each hidden layer.
    - activation : `torch.nn.Module`, default `torch.nn.ReLU()`
        Activation module of the hidden layers.
    - out_activation : `torch.nn.Module`, default `torch.nn.Identity()`
        Activation module of the output layer.

    Shape
    -----
    - Input : (N, *, H_in)
        where where * means any number of
        additional dimensions and `H_in = in_features`.
    - Output : (N, *, H_out)
        where all but the last dimension
        are the same shape as the input and `H_out = out_features`.

    Examples
    --------
    >>> _ = torch.manual_seed(42)
    >>> m = MultiLayerPerceptron()
    >>> _ = m(torch.empty((1, 2)))  # lazily determine number of input features
    >>> m
    MultiLayerPerceptron(
      (0): Linear(in_features=2, out_features=32, bias=True)
      (1): ReLU()
      (2): Linear(in_features=32, out_features=32, bias=True)
      (3): ReLU()
      (4): Linear(in_features=32, out_features=32, bias=True)
      (5): ReLU()
      (6): Linear(in_features=32, out_features=32, bias=True)
      (7): ReLU()
      (8): Linear(in_features=32, out_features=1, bias=True)
      (9): Identity()
    )
    >>> m(torch.randn((3, 2)))
    tensor([[...],
            [...],
            [...]], grad_fn=<AddmmBackward>)
    """

    def __init__(
        self,
        out_features=1,
        n_layers=4,
        n_units=32,
        activation=ReLU(),
        out_activation=Identity(),
    ):
        super().__init__()

        for _ in range(n_layers):
            self.append(LazyLinear(n_units))
            self.append(deepcopy(activation))
        self.append(LazyLinear(out_features))
        self.append(deepcopy(out_activation))

    def forward(self, input):
        x = input
        for layer in self:
            x = layer(x)
        return x



from .bs.american_binary import BSAmericanBinaryOption
from .bs.bs import BlackScholes
from .bs.european import BSEuropeanOption
from .bs.european_binary import BSEuropeanBinaryOption
from .bs.lookback import BSLookbackOption
from .modules.clamp import Clamp
from .modules.clamp import LeakyClamp
from .modules.loss import EntropicLoss
from .modules.loss import EntropicRiskMeasure
from .modules.loss import ExpectedShortfall
from .modules.loss import IsoelasticLoss
from .modules.mlp import MultiLayerPerceptron
from .modules.naked import Naked
from .modules.ww import WhalleyWilmott



import torch


def bisect(
    function, target, lower, upper, precision=1e-6, abort=100000
) -> torch.Tensor:
    """
    Perform binary search over a tensor.

    The output tensor approximately satisfies the following relation:

        function(output) = target

    Parameters
    ----------
    - function : callable
        Monotone increasing or decreasing function.
    - target : Tensor
        Target of function values.
    - lower : Tensor | float
        Lower bound of binary search.
    - upper : Tensor | float
        Upper bound of binary search.
    - precision : float, default 1e-6
        Precision of output.
    - abort : int, default 100000
        If the number of iterations exceeds this value, abort computation.

    Returns
    -------
    output : Tensor
        Arguments of functions that satisfy `f(output) = target`.

    Raises
    ------
    RuntimeError
        If the number of iteration exceeds `abort`.

    Examples
    --------
    >>> target = torch.tensor([-1.0, 0.0, 1.0])
    >>> function = torch.log
    >>> output = bisect(function, target, 0.01, 10.0)
    >>> output
    tensor([0.3679, 1.0000, 2.7183])
    >>> torch.allclose(function(output), target, atol=1e-6)
    True

    Monotone decreasing function:
    >>> function = lambda x: -torch.log(x)
    >>> output = bisect(function, target, 0.01, 10.0)
    >>> output
    tensor([2.7183, 1.0000, 0.3679])
    >>> torch.allclose(function(output), target, atol=1e-6)
    True
    """
    lower, upper = map(torch.as_tensor, (lower, upper))

    if not (lower < upper).all():
        raise ValueError("condition lower < upper should be satisfied.")

    if (function(lower) > function(upper)).all():
        # If decreasing function
        mf = lambda x: -function(x)
        return bisect(mf, -target, lower, upper, precision=precision, abort=abort)

    lower = torch.full_like(target, lower)
    upper = torch.full_like(target, upper)

    n_iter = 0
    while torch.max(upper - lower) > precision:
        n_iter += 1
        if n_iter > abort:
            raise RuntimeError(f"Aborting since iteration exceeds abort={abort}.")

        m = (lower + upper) / 2
        output = function(m)
        lower = torch.where(output < target, m, lower)
        upper = torch.where(output >= target, m, upper)

    return upper



from math import ceil

import torch
import torch.nn.functional as fn


def european_payoff(input: torch.Tensor, call=True, strike=1.0) -> torch.Tensor:
    """
    Returns the payoff of a European option.

    Parameters
    ----------
    - input : Tensor, shape (N_STEPS, *)
        The input tensor representing the price trajectory.
        Here, `N_STEPS` stands for the number of time steps and
        `*` means any number of additional dimensions.
    - call : bool, default True
        Specify whether the option is call or put.
    - strike : float, default 1.0
        The strike price of the option.

    Returns
    -------
    payoff : Tensor, shape (*)
    """
    if call:
        return fn.relu(input[-1] - strike)
    else:
        return fn.relu(strike - input[-1])


def lookback_payoff(input: torch.Tensor, call=True, strike=1.0):
    """
    Returns the payoff of a lookback option with a fixed strike.

    Parameters
    ----------
    - input : Tensor, shape (N_STEPS, *)
        The input tensor representing the price trajectory.
        Here, `N_STEPS` stands for the number of time steps and
        `*` means any number of additional dimensions.
    - call : bool, default True
        Specify whether the option is call or put.
    - strike : float, default 1.0
        The strike price of the option.

    Returns
    -------
    payoff : Tensor, shape (*)
    """
    if call:
        return fn.relu(input.max(dim=0).values - strike)
    else:
        return fn.relu(strike - input.min(dim=0).values)


def american_binary_payoff(input: torch.Tensor, call=True, strike=1.0):
    """
    Returns the payoff of an American binary option.

    Parameters
    ----------
    - input : Tensor, shape (N_STEPS, *)
        The input tensor representing the price trajectory.
        Here, `N_STEPS` stands for the number of time steps and
        `*` means any number of additional dimensions.
    - call : bool, default True
        Specify whether the option is call or put.
    - strike : float, default 1.0
        The strike price of the option.

    Returns
    -------
    payoff : Tensor, shape (*)
    """
    if call:
        return (input.max(dim=0).values >= strike).to(input)
    else:
        return (input.min(dim=0).values <= strike).to(input)


def european_binary_payoff(input: torch.Tensor, call=True, strike=1.0):
    """
    Returns the payoff of a European binary option.

    Parameters
    ----------
    - input : Tensor, shape (N_STEPS, *)
        The input tensor representing the price trajectory.
        Here, `N_STEPS` stands for the number of time steps and
        `*` means any number of additional dimensions.
    - call : bool, default True
        Specify whether the option is call or put.
    - strike : float, default 1.0
        The strike price of the option.

    Returns
    -------
    payoff : Tensor, shape (*)
    """
    if call:
        return (input[-1] >= strike).to(input)
    else:
        return (input[-1] <= strike).to(input)


def exp_utility(input: torch.Tensor, a=1.0) -> torch.Tensor:
    """
    Applies an exponential utility function.

    An exponential utility function is defined as:

        u(input) = -exp(-a * input)

    Parameters
    ----------
    - input : Tensor, shape (*)
        The input tensor.
    - a : float, default 1.0
        The risk aversion coefficient of the exponential utility.

    Returns
    -------
    exp_utility : Tensor, shape (*)
    """
    return -torch.exp(-a * input)


def isoelastic_utility(input: torch.Tensor, a=0.5):
    """
    Applies an isoelastic utility function.

    An isoelastic utility function is defined as:

        u(x) = x ** (1 - a) if a != 1.0
               log(x) if a == 1.0

    Parameters
    ----------
    - input : Tensor, shape (*)
        The input tensor.
    - a : float, default 1.0
        Relative risk aversion coefficient of the isoelastic utility.

    Returns
    -------
    isoelastic_utility : Tensor, shape (*)
    """
    if a == 1.0:
        return torch.log(input)
    else:
        return torch.pow(input, exponent=1.0 - a)


def topp(input, p, dim=None, largest=True) -> (torch.Tensor, torch.LongTensor):
    """
    Returns the largest `p * N` elements of the given input tensor,
    where `N` stands for the total number of elements in the input tensor.

    If `largest` is `False` then the smallest elements are returned.

    Parameters
    ----------
    - input : Tensor
        The input tensor.
    - p : float
        Quantile level.
    - dim : int, optional
        The dimension to sort along.
    - largest : bool, default True
        Controls whether to return largest or smallest elements.

    Examples
    --------
    >>> input = torch.arange(1., 6.)
    >>> input
    tensor([1., 2., 3., 4., 5.])
    >>> topp(input, 3 / 5)
    torch.return_types.topk(
    values=tensor([5., 4., 3.]),
    indices=tensor([4, 3, 2]))
    """
    if dim is None:
        return torch.topk(input, ceil(p * input.numel()), largest=largest)
    else:
        return torch.topk(input, ceil(p * input.size()[dim]), dim=dim, largest=largest)


def expected_shortfall(input: torch.Tensor, p: float, dim=None) -> torch.Tensor:
    """
    Returns the expected shortfall of the given input tensor.

    Parameters
    ----------
    - input : Tensor, shape (*)
        Input tensor.
    - p : float
        Quantile level.
    - dim : int
        The dimension or dimensions to reduce.

    Examples
    --------
    >>> input = -torch.arange(1., 6.)
    >>> expected_shortfall(input, 3 / 5)
    tensor(4.)

    Returns
    -------
    expected_shortfall : Tensor
    """
    if dim is None:
        return -topp(input, p=p, largest=False).values.mean()
    else:
        return -topp(input, p=p, largest=False, dim=dim).values.mean(dim=dim)


def leaky_clamp(
    input, min_value=None, max_value=None, clamped_slope=0.01
) -> torch.Tensor:
    """
    Clamp all elements in the input tensor into the range [`min_value`, `max_value`]
    and return a resulting tensor.
    The bounds `min_value` and `max_value` can be tensors.

    See `pfhedge.nn.LeakyClamp` for details.
    """
    x = input

    if min_value is not None:
        min_value = torch.as_tensor(min_value)
        x = torch.max(input, min_value + clamped_slope * (x - min_value))

    if max_value is not None:
        max_value = torch.as_tensor(max_value)
        x = torch.min(x, max_value + clamped_slope * (x - max_value))

    if min_value is not None and max_value is not None:
        x = torch.where(min_value <= max_value, x, (min_value + max_value) / 2)

    return x


def clamp(input, min_value=None, max_value=None) -> torch.Tensor:
    """
    Clamp all elements in the input tensor into the range [`min_value`, `max_value`]
    and return a resulting tensor.
    The bounds `min_value` and `max_value` can be tensors.

    See `pfhedge.nn.Clamp` for details.
    """
    return leaky_clamp(
        input, min_value=min_value, max_value=max_value, clamped_slope=0.0
    )



from ..nn.functional import lookback_payoff
from ._base import Derivative


class LookbackOption(Derivative):
    """
    A lookback option with fixed strike.

    A lookback call option provides its holder the right to buy an underlying with
    the strike price and to sell with the highest price until the date of maturity.

    A lookback put option provides its holder the right to sell an underlying with
    the strike price and to buy with the lowest price until the date of maturity.

    The payoff of a lookback call option is given by:

        payoff = max(Max - K, 0)

        Max = The maximum of the underlying asset's price until maturity
        K = The strike price (`strike`) of the option

    The payoff of a lookback call option is given by:

        payoff = max(K - Min, 0)

        Min = The minimum of the underlying asset's price until maturity
        K = The strike price (`strike`) of the option

    Parameters
    ----------
    - underlier : Primary
        The underlying instrument of the option.
    - call : bool, default True
        Specify whether the option is call or put.
    - strike : float, default 1.0
        The strike price of the option.
    - maturity : float, default 30 / 365
        The maturity of the option.

    Examples
    --------
    >>> import torch
    >>> from pfhedge.instruments import BrownianStock

    >>> _ = torch.manual_seed(42)
    >>> deriv = LookbackOption(BrownianStock(), maturity=5 / 365)
    >>> deriv.simulate(n_paths=2)
    >>> deriv.underlier.prices
    tensor([[1.0000, 1.0000],
            [1.0024, 1.0024],
            [0.9906, 1.0004],
            [1.0137, 0.9936],
            [1.0186, 0.9964]])
    >>> deriv.payoff()
    tensor([0.0186, 0.0024])
    """

    def __init__(self, underlier, call=True, strike=1.0, maturity=30 / 365):
        super().__init__()
        self.underlier = underlier
        self.call = call
        self.strike = strike
        self.maturity = maturity

    def __repr__(self):
        params = [f"{self.underlier.__class__.__name__}(...)"]
        if not self.call:
            params.append(f"call={self.call}")
        if self.strike != 1.0:
            params.append(f"strike={self.strike}")
        params.append(f"maturity={self.maturity:.2e}")

        return self.__class__.__name__ + f"({', '.join(params)})"

    def payoff(self):
        return lookback_payoff(
            self.underlier.prices, call=self.call, strike=self.strike
        )



from ._getter import FEATURES
from ._getter import get_feature
from .features import Barrier
from .features import ExpiryTime
from .features import LogMoneyness
from .features import MaxLogMoneyness
from .features import MaxMoneyness
from .features import ModuleOutput
from .features import Moneyness
from .features import PrevHedge
from .features import Volatility
from .features import Zero



from ..nn.functional import european_payoff
from ._base import Derivative


class EuropeanOption(Derivative):
    """
    A European option.

    A European option provides its holder the right to buy (for call option)
    or sell (for put option) an underlying asset with the strike price
    on the date of maturity.

    The payoff of a European call option is given by:

        payoff = max(S - K, 0)

        S = The underlying asset's price at maturity
        K = The strike price (`strike`) of the option

    The payoff of a European put option is given by:

        payoff = max(K - S, 0)

        S = The underlying asset's price at maturity
        K = The strike price (`strike`) of the option

    Parameters
    ----------
    - underlier : Primary
        The underlying instrument of the option.
    - call : bool, default True
        Specify whether the option is call or put.
    - strike : float, default 1.0
        The strike price of the option.
    - maturity : float, default 30 / 365
        The maturity of the option.

    Examples
    --------
    >>> import torch
    >>> from pfhedge.instruments import BrownianStock

    >>> _ = torch.manual_seed(42)
    >>> deriv = EuropeanOption(BrownianStock(), maturity=5 / 365)
    >>> deriv.simulate(n_paths=2)
    >>> deriv.underlier.prices
    tensor([[1.0000, 1.0000],
            [1.0024, 1.0024],
            [0.9906, 1.0004],
            [1.0137, 0.9936],
            [1.0186, 0.9964]])
    >>> deriv.payoff()
    tensor([0.0186, 0.0000])
    """

    def __init__(self, underlier, call=True, strike=1.0, maturity=30 / 365):
        super().__init__()
        self.underlier = underlier
        self.call = call
        self.strike = strike
        self.maturity = maturity

    def __repr__(self):
        params = [f"{self.underlier.__class__.__name__}(...)"]
        if not self.call:
            params.append(f"call={self.call}")
        if self.strike != 1.0:
            params.append(f"strike={self.strike}")
        params.append(f"maturity={self.maturity:.2e}")

        return self.__class__.__name__ + f"({', '.join(params)})"

    def payoff(self):
        return european_payoff(
            self.underlier.prices, call=self.call, strike=self.strike
        )



from ._base import Feature
from .features import ExpiryTime
from .features import LogMoneyness
from .features import MaxLogMoneyness
from .features import MaxMoneyness
from .features import Moneyness
from .features import PrevHedge
from .features import Volatility
from .features import Zero

FEATURES = [
    ExpiryTime(),
    LogMoneyness(),
    MaxLogMoneyness(),
    MaxMoneyness(),
    Moneyness(),
    PrevHedge(),
    Volatility(),
    Zero(),
]


def get_feature(feature):
    """
    Get feature from name.

    Parameters
    ----------
    - name : str or Feature
        Name of feature.

    Returns
    -------
    feature : Feature
        Feature.
    """
    dict_features = {str(f): f for f in FEATURES}

    if isinstance(feature, str):
        f = dict_features.get(feature)
        if f is None:
            raise ValueError(
                f"{feature} is not a valid value. "
                "Use sorted(pfhedge.features.FEATURES) to get valid options."
            )
    else:
        f = feature
        if not isinstance(feature, Feature):
            raise TypeError(f"{feature} is not an instance of Feature.")
    return f



import torch

from ..bs.bs import BlackScholes
from .clamp import Clamp


class WhalleyWilmott(torch.nn.Module):
    """
    Initialize Whalley-Wilmott's hedging strategy of a derivative.
    The `forward` method returns the next hedge ratio.

    This is the optimal hedging strategy for asymptotically small transaction cost.

    See:
    Whalley, A.E. and Wilmott, P., An asymptotic analysis of an optimal hedging model
    for option pricing with transaction costs. Mathematical Finance, 1997, 7, 307–324.

    Parameters
    ----------
    - derivative : Derivative
        Derivative to hedge.
    - a : float, default 1.0
        Risk aversion parameter in exponential utility.

    Shape
    -----
    Input : (N, *, H_in)
        See `features()` for input features.
        Here, `*` means any number of additional dimensions and `H_in` is
        the number of input features.
    Output : (N, *, 1)
        The hedge ratio at the next time step.

    Examples
    --------
    >>> from pfhedge.instruments import BrownianStock
    >>> from pfhedge.instruments import EuropeanOption
    >>> from pfhedge.instruments import LookbackOption

    European option:
    >>> deriv = EuropeanOption(BrownianStock(cost=1e-5))
    >>> m = WhalleyWilmott(deriv)
    >>> m.features()
    ['log_moneyness', 'expiry_time', 'volatility', 'prev_hedge']
    >>> input = torch.tensor([
    ...     [-0.05, 0.1, 0.2, 0.5],
    ...     [-0.01, 0.1, 0.2, 0.5],
    ...     [ 0.00, 0.1, 0.2, 0.5],
    ...     [ 0.01, 0.1, 0.2, 0.5],
    ...     [ 0.05, 0.1, 0.2, 0.5]])
    >>> m(input)
    tensor([[0.2946],
            [0.5000],
            [0.5000],
            [0.5000],
            [0.7284]])

    European option with zero cost:
    >>> deriv = EuropeanOption(BrownianStock())
    >>> m = WhalleyWilmott(deriv)
    >>> m.features()
    ['log_moneyness', 'expiry_time', 'volatility', 'prev_hedge']
    >>> input = torch.tensor([
    ...     [-0.05, 0.1, 0.2, 0.5],
    ...     [-0.01, 0.1, 0.2, 0.5],
    ...     [ 0.00, 0.1, 0.2, 0.5],
    ...     [ 0.01, 0.1, 0.2, 0.5],
    ...     [ 0.05, 0.1, 0.2, 0.5]])
    >>> m(input)
    tensor([[0.2239],
            [0.4497],
            [0.5126],
            [0.5752],
            [0.7945]])
    """

    def __init__(self, derivative, a=1.0):
        super().__init__()
        self.derivative = derivative
        self.a = a

        self.bs = BlackScholes(derivative)
        self.clamp = Clamp()

    def features(self):
        return self.bs.features() + ["prev_hedge"]

    def extra_repr(self):
        return f"a={self.a}" if self.a != 1 else ""

    def forward(self, input) -> torch.Tensor:
        prev_hedge = input[..., [-1]]

        delta = self.bs(input[..., :-1])
        width = self.width(input[..., :-1])

        return self.clamp(prev_hedge, min_value=delta - width, max_value=delta + width)

    def width(self, input) -> torch.Tensor:
        """
        Return half-width of the no-transaction band.

        Parameters
        ----------
        - input : Tensor, shape (N, *, H_in - 1)

        Returns
        -------
        width : Tensor, shape (N, *, 1)
        """
        cost = self.derivative.underlier.cost

        price = self.derivative.strike * torch.exp(input[..., [0]])
        gamma = self.bs.gamma(*(input[..., [i]] for i in range(input.size()[-1])))
        width = (cost * (3 / 2) * (gamma ** 2) * price / self.a) ** (1 / 3)

        return width



import torch

from ..._utils.bisect import bisect
from ._base import BSModuleMixin


class BSEuropeanOption(BSModuleMixin):
    """
    Black-Scholes formula for a European option.

    Parameters
    ----------
    - derivative : EuropeanOption, optional
        The derivative to get the Black-Scholes formula.
    - call : bool, default True
        Specify whether the option is call or put.
    - strike : float, default 1.0
        The strike price of the option.

    Shape
    -----
    - Input : (N, *, 3)
        See `features()` for input features.
        Here, `*` means any number of additional dimensions.
    - Output : (N, *, 1)
        Delta of the derivative.
        All but the last dimension are the same shape as the input.

    Examples
    --------
    >>> m = BSEuropeanOption()
    >>> m.features()
    ['log_moneyness', 'expiry_time', 'volatility']
    >>> x = torch.tensor([
    ...     [-0.01, 0.1, 0.2],
    ...     [ 0.00, 0.1, 0.2],
    ...     [ 0.01, 0.1, 0.2]])
    >>> m(x)
    tensor([[0.4497],
            [0.5126],
            [0.5752]])
    """

    def __init__(self, derivative=None, call=True, strike=1.0):
        super().__init__()

        if derivative is not None:
            self.call = derivative.call
            self.strike = derivative.strike
        else:
            self.call = call
            self.strike = strike

    def extra_repr(self):
        params = []
        if not self.call:
            params.append(f"call={self.call}")
        if self.strike != 1.0:
            params.append(f"strike={self.strike}")
        return ", ".join(params)

    def delta(self, log_moneyness, expiry_time, volatility) -> torch.Tensor:
        """
        Returns delta of the derivative.

        Parameters
        ----------
        - log_moneyness : Tensor, shape (N, *)
            Log moneyness of the prices of the underlying asset.
        - expiry_time : Tensor, shape (N, *)
            Time to expiry of the option.
        - volatility : Tensor, shape (N, *)
            Volatility of the underlying asset.

        Returns
        -------
        delta : Tensor, shape (N, *)
        """
        s, t, v = map(torch.as_tensor, (log_moneyness, expiry_time, volatility))

        delta = self.N.cdf(self.d1(s, t, v))
        delta = delta - 1 if not self.call else delta  # put-call parity

        return delta

    def gamma(self, log_moneyness, expiry_time, volatility) -> torch.Tensor:
        """
        Returns gamma of the derivative.

        Parameters
        ----------
        - log_moneyness : Tensor, shape (N, *)
            Log moneyness of the prices of the underlying asset.
        - expiry_time : Tensor, shape (N, *)
            Time to expiry of the option.
        - volatility : Tensor, shape (N, *)
            Volatility of the underlying asset.

        Returns
        -------
        gamma : Tensor, shape (N, *)
        """
        if not self.call:
            raise ValueError(
                f"{self.__class__.__name__} for a put option is not yet supported."
            )

        s, t, v = map(torch.as_tensor, (log_moneyness, expiry_time, volatility))
        price = self.strike * torch.exp(s)
        gamma = self.N.pdf(self.d1(s, t, v)) / (price * v * torch.sqrt(t))

        return gamma

    def price(self, log_moneyness, expiry_time, volatility) -> torch.Tensor:
        """
        Returns price of the derivative.

        Parameters
        ----------
        - log_moneyness : Tensor, shape (N, *)
            Log moneyness of the prices of the underlying asset.
        - max_log_moneyness : Tensor, shape (N, *)
            Cumulative maximum of the log moneyness.
        - expiry_time : Tensor, shape (N, *)
            Time to expiry of the option.
        - volatility : Tensor, shape (N, *)
            Volatility of the underlying asset.

        Returns
        -------
        price : Tensor, shape (N, *)
        """
        s, t, v = map(torch.as_tensor, (log_moneyness, expiry_time, volatility))

        n1 = self.N.cdf(self.d1(s, t, v))
        n2 = self.N.cdf(self.d2(s, t, v))

        price = self.strike * (torch.exp(s) * n1 - n2)

        if not self.call:
            price += self.strike * (torch.exp(s) - 1)  # put-call parity

        return price

    def implied_volatility(
        self, log_moneyness, expiry_time, price, precision=1e-6
    ) -> torch.Tensor:
        """
        Returns implied volatility of the derivative.

        Parameters
        ----------
        - log_moneyness : Tensor, shape (N, *)
            Log moneyness of the prices of the underlying asset.
        - expiry_time : Tensor, shape (N, *)
            Time to expiry of the option.
        - price : Tensor, shape (N, *)
            Price of the derivative.
        - precision : float, default 1e-6
            Computational precision of the implied volatility.

        Returns
        -------
        implied_volatility : Tensor, shape (N, *)
        """
        s, t, p = map(torch.as_tensor, (log_moneyness, expiry_time, price))
        get_price = lambda v: self.price(s, t, v)
        return bisect(get_price, p, lower=0.001, upper=1.000, precision=precision)



import abc
from inspect import signature

import torch


class _Normal(torch.distributions.normal.Normal):
    """
    Creates a normal (also called Gaussian) distribution parameterized
    by `loc` and `scale`.
    It has `pdf` method to compute the probability density function.

    See `torch.distributions.normal.Normal` for details and other methods.
    """

    def pdf(self, value: torch.Tensor) -> torch.Tensor:
        """
        Returns the probability density function evaluated at value.

        Parameters
        ----------
        - value : Tensor, shape (*)

        Returns
        -------
        pdf : Tensor, shape (*)
        """
        return self.log_prob(value).exp()


class BSModuleMixin(torch.nn.Module):
    """
    A mixin class for Black-Scholes formula modules.

    Shape
    -----
    - Input : (N, *, H_in)
        See `features()` for input features.
        Here, `*` means any number of additional dimensions.
    - Output : (N, *, 1)
        Delta of the derivative.
        All but the last dimension are the same shape as the input.
    """

    def forward(self, input) -> torch.Tensor:
        """
        Returns delta of the derivative.

        Parameters
        ----------
        - input : torch.Tensor, shape (N, *, H_in)
            The input tensor.
            Features are concatenated along the last dimension.

        Returns
        -------
        output : torch.Tensor, shape (N, *, 1)
        """
        return self.delta(*(input[..., [i]] for i in range(input.size()[-1])))

    @abc.abstractmethod
    def delta(*args, **kwargs) -> torch.Tensor:
        """
        Returns delta of the derivative.

        Returns
        -------
        delta : Tensor, shape (N, *)
        """

    def features(self) -> list:
        """
        Returns a list of names of input features.

        By default, this method infers the names of features from the signature of
        the `delta` method.
        Please override this method if the signature is different from
        the feature names.
        """
        return list(signature(self.delta).parameters.keys())

    @property
    def N(self):
        """
        Returns normal distribution with zero mean and unit standard deviation.
        """
        return _Normal(torch.tensor(0.0), torch.tensor(1.0))

    @staticmethod
    def d1(log_moneyness, expiry_time, volatility) -> torch.Tensor:
        """
        Returns `d1` in the Black-Scholes formula.

        Parameters
        ----------
        - log_moneyness : Tensor, shape (*)
            Log moneyness of the underlying asset.
        - expiry_time : Tensor, shape (*)
            Time to expiry of the option.
        - volatility : Tensor, shape (*)
            Volatility of the underlying asset.

        Returns
        -------
        d1 : Tensor, shape (*)
        """
        s, t, v = map(torch.as_tensor, (log_moneyness, expiry_time, volatility))
        return (s + (v ** 2 / 2) * t) / (v * torch.sqrt(t))

    @staticmethod
    def d2(log_moneyness, expiry_time, volatility) -> torch.Tensor:
        """
        Returns `d2` in the Black-Scholes formula.

        Parameters
        ----------
        - log_moneyness : Tensor, shape (*)
            Log moneyness of the underlying asset.
        - expiry_time : Tensor, shape (*)
            Time to expiry of the option.
        - volatility : Tensor, shape (*)
            Volatility of the underlying asset.

        Returns
        -------
        d2 : Tensor, shape (*)
        """
        s, t, v = map(torch.as_tensor, (log_moneyness, expiry_time, volatility))
        return (s - (v ** 2 / 2) * t) / (v * torch.sqrt(t))



import torch

from ._base import Feature


class Moneyness(Feature):
    """
    Moneyness of the underlying instrument of the derivative.

    Parameters
    ----------
    - log : bool, default False
        If `True`, represents log moneyness.
    """

    def __init__(self, log=False):
        super().__init__()

        self.log = log

    def __str__(self):
        if self.log:
            return "log_moneyness"
        else:
            return "moneyness"

    def __getitem__(self, i):
        s = self.derivative.underlier.prices[i].reshape(-1, 1)
        output = s / self.derivative.strike
        if self.log:
            output = torch.log(output)
        return output


class LogMoneyness(Moneyness):
    """
    Log moneyness of the underlying instrument of the derivative.
    """

    def __init__(self):
        super().__init__(log=True)


class ExpiryTime(Feature):
    """
    Remaining time to the maturity of the derivative.
    """

    def __str__(self):
        return "expiry_time"

    def __getitem__(self, i):
        value = self.derivative.maturity - i * self.derivative.underlier.dt
        return torch.full_like(self.derivative.underlier.prices.T[:, :1], value)


class Volatility(Feature):
    """
    Volatility of the underlier of the derivative.
    """

    def __str__(self):
        return "volatility"

    def __getitem__(self, i):
        value = self.derivative.underlier.volatility
        return torch.full_like(self.derivative.underlier.prices.T[:, :1], value)


class PrevHedge(Feature):
    """
    Previous holding of underlier.
    """

    def __str__(self):
        return "prev_hedge"

    def __getitem__(self, i):
        if hasattr(self.hedger, "prev"):
            return self.hedger.prev.reshape(-1, 1)
        else:
            return torch.zeros_like(self.derivative.underlier.prices.T[:, :1])


class Barrier(Feature):
    """
    A feature which signifies whether the price of the underlier have reached
    the barrier. The output `1.0` means that the price have touched the barrier,
    and `0` otherwise.

    Parameters
    ----------
    - barrier : float
        The price level of the barrier.
    - up : bool, default True
        If `True`, signifies whether the price has exceeded the barrier upward.
        If `False`, signifies whether the price has exceeded the barrier downward.
    """

    def __init__(self, barrier, up=True):
        super().__init__()

        self.barrier = barrier
        self.up = up

    def __repr__(self):
        return self.__class__.__name__ + f"({self.barrier}, up={self.up})"

    def __getitem__(self, i):
        if self.up:
            touch_barrier = self.derivative.underlier.prices[: i + 1] >= self.barrier
        else:
            touch_barrier = self.derivative.underlier.prices[: i + 1] <= self.barrier
        return (
            touch_barrier.any(dim=0).reshape(-1, 1).to(self.derivative.underlier.prices)
        )


class Zero(Feature):
    """
    A feature of which value is always zero.
    """

    def __str__(self):
        return "zero"

    def __getitem__(self, i):
        return torch.zeros_like(self.derivative.underlier.prices.T[:, :1])


class MaxMoneyness(Feature):
    """
    Cumulative maximum of moneyness.

    Parameters
    ----------
    - log : bool, default False
        If `True`, represents log moneyness.
    """

    def __init__(self, log=False):
        self.log = log
        self.moneyness = Moneyness(log=log)

    def __str__(self):
        if self.log:
            return "max_log_moneyness"
        else:
            return "max_moneyness"

    def __getitem__(self, i):
        s = self.derivative.underlier.prices[: i + 1].max(dim=0).values.reshape(-1, 1)
        output = s / self.derivative.strike
        if self.log:
            output = torch.log(output)
        return output

    def of(self, derivative=None, hedger=None):
        super().of(derivative)
        self.moneyness.of(derivative=derivative, hedger=hedger)
        return self


class MaxLogMoneyness(MaxMoneyness):
    """
    Cumulative maximum of log Moneyness.
    """

    def __init__(self):
        super().__init__(log=True)


class ModuleOutput(Feature, torch.nn.Module):
    """
    The feature computed as an output of a `torch.nn.Module`.

    Parameters
    ----------
    - module : torch.nn.Module
        Module to compute the value of the feature.
        The input and output shapes should be `(N, *, H_in) -> (N, *, 1)`,
        where `N` stands for the number of Monte Carlo paths of the underlier of
        the derivative, `H_in` stands for the number of input features
        (namely, `H_in = len(features)`),
        and `*` means any number of additional dimensions.
    - features : list[Feature]
        The input features to the module.

    Examples
    --------
    >>> from torch.nn import Linear
    >>> from pfhedge.instruments import BrownianStock
    >>> from pfhedge.instruments import EuropeanOption

    >>> _ = torch.manual_seed(42)
    >>> deriv = EuropeanOption(BrownianStock())
    >>> deriv.simulate(n_paths=3)
    >>> m = Linear(2, 1)
    >>> f = ModuleOutput(m, [Moneyness(), ExpiryTime()]).of(deriv)
    >>> f[0]
    tensor([[...],
            [...],
            [...]], grad_fn=<AddmmBackward>)
    >>> f
    ModuleOutput(
      features=['moneyness', 'expiry_time'],
      (module): Linear(in_features=2, out_features=1, bias=True)
    )

    >>> from pfhedge.nn import BlackScholes

    >>> _ = torch.manual_seed(42)
    >>> deriv = EuropeanOption(BrownianStock())
    >>> deriv.simulate(n_paths=3)
    >>> m = BlackScholes(deriv)
    >>> f = ModuleOutput(m, [LogMoneyness(), ExpiryTime(), Volatility()]).of(deriv)
    >>> f[0]
    tensor([[0.5114],
            [0.5114],
            [0.5114]])
    """

    def __init__(self, module, features):
        super().__init__()

        self.module = module
        self.features = features

    def extra_repr(self):
        return f"features={[str(f) for f in self.features]},"

    def forward(self, input):
        return self.module(input)

    def __getitem__(self, i):
        return self(torch.cat([f[i] for f in self.features], 1))

    def of(self, derivative=None, hedger=None):
        super().of(derivative, hedger)
        self.features = [feature.of(derivative, hedger) for feature in self.features]
        return self



import torch

from ..._utils.bisect import bisect
from ._base import BSModuleMixin


class BSAmericanBinaryOption(BSModuleMixin):
    """
    Black-Scholes formulas for American Binary Option.

    Parameters
    ----------
    - derivative : AmericanBinaryOption
        The derivative to get the Black-Scholes formula.
    - call : bool, default True
        Specify whether the option is call or put.
    - strike : float, default 1.0
        The strike price of the option.

    Shape
    -----
    - Input : (N, *, 4)
        See `features()` for input features.
        Here, `*` means any number of additional dimensions.
    - Output : (N, *, 1)
        Delta of the derivative.
        All but the last dimension are the same shape as the input.

    Examples
    --------
    >>> m = BSAmericanBinaryOption(strike=1.0)
    >>> m.features()
    ['log_moneyness', 'max_log_moneyness', 'expiry_time', 'volatility']
    >>> x = torch.tensor([
    ...     [-0.01, -0.01, 0.1, 0.2],
    ...     [ 0.00,  0.00, 0.1, 0.2],
    ...     [ 0.01,  0.01, 0.1, 0.2]])
    >>> m(x)
    tensor([[1.1285],
            [0.0000],
            [0.0000]])
    """

    def __init__(self, derivative=None, call=True, strike=1.0):
        super().__init__()

        if derivative is not None:
            self.call = derivative.call
            self.strike = derivative.strike
        else:
            self.call = call
            self.strike = strike

        if not self.call:
            raise ValueError(
                f"{self.__class__.__name__} for a put option is not yet supported."
            )

    def extra_repr(self):
        params = []
        if self.strike != 1.0:
            params.append(f"strike={self.strike}")
        return ", ".join(params)

    def delta(
        self, log_moneyness, max_log_moneyness, expiry_time, volatility
    ) -> torch.Tensor:
        """
        Returns delta of the derivative.

        Parameters
        ----------
        - log_moneyness : Tensor, shape (N, *)
            Log moneyness of the prices of the underlying asset.
        - max_log_moneyness : Tensor, shape (N, *)
            Cumulative maximum of the log moneyness.
        - expiry_time : Tensor, shape (N, *)
            Time to expiry of the option.
        - volatility : Tensor, shape (N, *)
            Volatility of the underlying asset.

        Returns
        -------
        delta : Tensor, shape (N, *)
        """
        s, m, t, v = map(
            torch.as_tensor, (log_moneyness, max_log_moneyness, expiry_time, volatility)
        )

        sqrt2 = torch.sqrt(torch.as_tensor(2.0)).item()
        d1 = self.d1(s, t, v)
        d2 = self.d2(s, t, v)
        c1 = self.N.cdf(d1 / sqrt2)
        p1 = self.N.pdf(d1 / sqrt2)
        p2 = self.N.pdf(d2 / sqrt2)

        d = (1 + c1 + (p1 + p2)) / (2 * self.strike)

        return torch.where(m < 0, d, torch.zeros_like(d))

    @torch.enable_grad()
    def gamma(
        self, log_moneyness, max_log_moneyness, expiry_time, volatility
    ) -> torch.Tensor:
        """
        Returns gamma of the derivative.

        Parameters
        ----------
        - log_moneyness : Tensor, shape (N, *)
            Log moneyness of the prices of the underlying asset.
        - max_log_moneyness : Tensor, shape (N, *)
            Cumulative maximum of the log moneyness.
        - expiry_time : Tensor, shape (N, *)
            Time to expiry of the option.
        - volatility : Tensor, shape (N, *)
            Volatility of the underlying asset.

        Returns
        -------
        gamma : Tensor, shape (N, *)
        """
        prices = self.strike * torch.exp(torch.as_tensor(log_moneyness))
        prices = torch.Tensor.requires_grad_(prices)

        s = torch.log(prices / self.strike)
        m, t, v = map(torch.as_tensor, (max_log_moneyness, expiry_time, volatility))
        m = torch.max(s, m)  # `s + epsilon` may be greater than `m`

        delta = self.delta(s, m, t, v)
        gamma = torch.autograd.grad(
            delta, prices, grad_outputs=torch.ones_like(prices)
        )[0]

        return gamma

    def price(
        self, log_moneyness, max_log_moneyness, expiry_time, volatility
    ) -> torch.Tensor:
        """
        Returns price of the derivative.

        Parameters
        ----------
        - log_moneyness : Tensor, shape (N, *)
            Log moneyness of the prices of the underlying asset.
        - max_log_moneyness : Tensor, shape (N, *)
            Cumulative maximum of the log moneyness.
        - expiry_time : Tensor, shape (N, *)
            Time to expiry of the option.
        - volatility : Tensor, shape (N, *)
            Volatility of the underlying asset.

        Returns
        -------
        price : Tensor, shape (N, *)
        """
        s, m, t, v = map(
            torch.as_tensor, (log_moneyness, max_log_moneyness, expiry_time, volatility)
        )

        sqrt2 = torch.sqrt(torch.as_tensor(2.0)).item()
        n1 = self.N.cdf(self.d1(s, t, v) / sqrt2)
        n2 = self.N.cdf(self.d2(s, t, v) / sqrt2)

        p = (1 / 2) * (torch.exp(s) * (1 + n1) + n2)

        return torch.where(m < 0, p, torch.ones_like(p))

    def implied_volatility(
        self, log_moneyness, max_log_moneyness, expiry_time, price, precision=1e-6
    ) -> torch.Tensor:
        """
        Returns implied volatility of the derivative.

        Parameters
        ----------
        - log_moneyness : Tensor, shape (N, *)
            Log moneyness of the prices of the underlying asset.
        - max_log_moneyness : Tensor, shape (N, *)
            Cumulative maximum of the log moneyness.
        - expiry_time : Tensor, shape (N, *)
            Time to expiry of the option.
        - price : Tensor, shape (N, *)
            Price of the derivative.
        - precision : float, default 1e-6
            Computational precision of the implied volatility.

        Returns
        -------
        implied_volatility : Tensor, shape (N, *)
        """
        s, m, t, p = map(
            torch.as_tensor, (log_moneyness, max_log_moneyness, expiry_time, price)
        )
        get_price = lambda volatility: self.price(s, m, t, volatility)
        return bisect(get_price, p, lower=0.001, upper=1.000, precision=precision)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pfhedge',
 'pfhedge._utils',
 'pfhedge.features',
 'pfhedge.instruments',
 'pfhedge.nn',
 'pfhedge.nn.bs',
 'pfhedge.nn.modules',
 'pfhedge.stochastic']

package_data = \
{'': ['*']}

install_requires = \
['torch>=1.8.0,<2.0.0', 'tqdm>=4.59.0,<5.0.0']

setup_kwargs = {
    'name': 'pfhedge',
    'version': '0.1.1',
    'description': 'Deep Hedging in PyTorch',
    'long_description': None,
    'author': 'Shota Imaki',
    'author_email': 'shota.imaki.0801@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/pfnet-research/pfhedge',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.13,<4.0.0',
}


setup(**setup_kwargs)



from .american_binary import AmericanBinaryOption
from .european import EuropeanOption
from .european_binary import EuropeanBinaryOption
from .lookback import LookbackOption
from .underlier import BrownianStock
